/*
 * This file is part of the Score-P software (http://www.score-p.org)
 *
 * Copyright (c) 2009-2012,
 * RWTH Aachen University, Germany
 *
 * Copyright (c) 2009-2012,
 * Gesellschaft fuer numerische Simulation mbH Braunschweig, Germany
 *
 * Copyright (c) 2009-2012, 2014,
 * Technische Universitaet Dresden, Germany
 *
 * Copyright (c) 2009-2012,
 * University of Oregon, Eugene, USA
 *
 * Copyright (c) 2009-2012,
 * Forschungszentrum Juelich GmbH, Germany
 *
 * Copyright (c) 2009-2012,
 * German Research School for Simulation Sciences GmbH, Juelich/Aachen, Germany
 *
 * Copyright (c) 2009-2012,
 * Technische Universitaet Muenchen, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */


/**
 *  @file
 *  @source     templates/otf2_attribute_list_inc.tmpl.c
 *
 *  @brief      Convenient helper function to add and get attributes.
 */

@otf2 for entry in Type.entries if entry.type:

OTF2_ErrorCode
OTF2_AttributeList_Add@@entry.type.otf2@@( OTF2_AttributeList*  attributeList,
                                           OTF2_AttributeRef    attributeID,
                                           @@entry.type@@       @@entry.type.lower@@Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.@@entry.type.lower@@ = @@entry.type.lower@@Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            @@entry.name@@,
                                            attribute_value );
}

@otf2 endfor
@otf2 for entry in MappingType.entries:
@otf2 with def = entry.def

OTF2_ErrorCode
OTF2_AttributeList_Add@@def.ref_name@@Ref( OTF2_AttributeList*    attributeList,
                                           OTF2_AttributeRef      attributeID,
                                           @@def.self_attr.type@@ @@def.ref_lname@@Ref )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.@@def.ref_lname@@Ref = @@def.ref_lname@@Ref;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            @@Type.prefix@@@@entry.suffix@@,
                                            attribute_value );
}

@otf2 endwith
@otf2 endfor
@otf2 for entry in Type.entries if entry.type:

OTF2_ErrorCode
OTF2_AttributeList_Get@@entry.type.otf2@@( const OTF2_AttributeList* attributeList,
                                           OTF2_AttributeRef         attributeID,
                                           @@entry.type@@*           @@entry.type.lower@@Value )
{
    if ( !@@entry.type.lower@@Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for @@entry.type.lower@@Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != @@entry.name@@ )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *@@entry.type.lower@@Value = attribute_value.@@entry.type.lower@@;

    return OTF2_SUCCESS;
}

@otf2 endfor
@otf2 for entry in MappingType.entries:
@otf2 with def = entry.def

OTF2_ErrorCode
OTF2_AttributeList_Get@@def.ref_name@@Ref( const OTF2_AttributeList*   attributeList,
                                           OTF2_AttributeRef           attributeID,
                                           @@def.self_attr.type@@*     @@def.ref_lname@@Ref )
{
    if ( !@@def.ref_lname@@Ref )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for @@def.ref_lname@@Ref." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != @@Type.prefix@@@@entry.suffix@@ )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *@@def.ref_lname@@Ref = attribute_value.@@def.ref_lname@@Ref;

    return OTF2_SUCCESS;
}

@otf2 endwith
@otf2 endfor

OTF2_ErrorCode
otf2_attribute_list_read_from_buffer( OTF2_AttributeList*  attributeList,
                                      OTF2_Buffer*         buffer,
                                      OTF2_MappingCallback mappingCb,
                                      void*                mappingData )
{
    UTILS_ASSERT( attributeList );
    UTILS_ASSERT( buffer );

    OTF2_ErrorCode status;

    /* Get record length and test memory availability */
    uint64_t record_data_length;
    status = OTF2_Buffer_GuaranteeRecord( buffer, &record_data_length );
    if ( OTF2_SUCCESS != status )
    {
        return UTILS_ERROR( status,
                               "Could not read attribute list. "
                               "Not enough memory in buffer." );
    }
    uint8_t* record_start_pos;
    OTF2_Buffer_GetPosition( buffer, &record_start_pos );

    /* Get how many attributes are in the list */
    uint32_t number_of_attributes;
    status = OTF2_Buffer_ReadUint32( buffer, &number_of_attributes );
    if ( OTF2_SUCCESS != status )
    {
        return UTILS_ERROR( status,
                            "Could not read number of attributes in list. "
                            "Invalid compression size." );
    }

    /* Check if number of attributes is consistent with record length */
    if ( record_data_length > number_of_attributes * OTF2_MAXIMUM_ATTRIBUTE_LIST_ENTRY_SIZE + 5 )
    {
        return UTILS_ERROR( OTF2_ERROR_INTEGRITY_FAULT,
                            "Could not read attribute list. "
                            "Record length and number of attributes in list are not consistent." );
    }

    for ( uint32_t i = 0; i < number_of_attributes; i++ )
    {
        uint32_t            attribute_id;
        OTF2_Type           type;
        OTF2_AttributeValue value;

        /* Read attribute id from trace */
        status = OTF2_Buffer_ReadUint32( buffer, &attribute_id );
        if ( OTF2_SUCCESS != status )
        {
            return UTILS_ERROR( status,
                                "Could not read attribute from list. "
                                "Invalid compression size." );
        }
        attribute_id = mappingCb( mappingData, OTF2_MAPPING_ATTRIBUTE, attribute_id );

        /* Read the type from trace */
        OTF2_Buffer_ReadUint8( buffer, &type );

        /* Cast the container to the right type and read it from the trace */
        switch ( type )
        {
            @otf2 for entry in Type.entries if entry.type:
            case @@entry.name@@:
                @otf2 if entry.type.errorcheck_on_read
                status = OTF2_Buffer_Read@@entry.type.otf2@@( buffer, &value.@@entry.type.lower@@ );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read attribute. "
                                        "@@entry.type.errorcheck_on_read@@" );
                }
                @otf2 else:
                OTF2_Buffer_Read@@entry.type.otf2@@( buffer, &value.@@entry.type.lower@@ );
                @otf2 endif
                break;

            @otf2 endfor
            @otf2 for entry in MappingType.entries:
            @otf2 with def = entry.def
            case @@Type.prefix@@@@entry.suffix@@:
                /* @@Type.prefix@@@@entry.suffix@@ attributes reference a @@def.siblings|join(', or a ')@@ definition */
                status = OTF2_Buffer_Read@@def.self_attr.type.otf2@@( buffer,
                                                                      &value.@@def.ref_lname@@Ref );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read @@def.name@@ attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.@@def.ref_lname@@Ref = mappingCb( mappingData,
                                                        @@entry@@,
                                                        value.@@def.ref_lname@@Ref );
                break;

            @otf2 endwith
            @otf2 endfor
            default:
                /*
                 * There can only be more definition references added in the future,
                 * and definition references are guaranteed to be of compressable
                 * size, thus we induce that the value is of compressed type
                 */
                status = OTF2_Buffer_ReadUint64( buffer, &value.uint64 );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read attribute of unknown type. "
                                        "Invalid compression size." );
                }
                break;
        }

        /* Store the new value in the attribute list. */
        OTF2_AttributeList_AddAttribute( attributeList,
                                         attribute_id,
                                         type,
                                         value );
    }

    /*
     * Always jump to the announced end of the record.
     */
    status = OTF2_Buffer_SetPosition( buffer, record_start_pos + record_data_length );
    if ( OTF2_SUCCESS != status )
    {
        return UTILS_ERROR( status, "Could not read attribute list." );
    }

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
otf2_attribute_list_write_to_buffer( OTF2_AttributeList* attributeList,
                                     OTF2_Buffer*        buffer )
{
    /* Write no attribute list to the buffer, if there is no one. */
    if ( attributeList == NULL || attributeList->capacity == 0 )
    {
        return OTF2_SUCCESS;
    }

    /* Write the attribute lists record type id to the trace */
    OTF2_Buffer_WriteUint8( buffer, OTF2_ATTRIBUTE_LIST );

    /* Write initial record length */
    uint32_t data_size_of_attr_list = otf2_attribute_list_get_data_size( attributeList );
    OTF2_Buffer_WriteInitialRecordLength( buffer, data_size_of_attr_list );

    /* Store how many items are in the list to the trace */
    OTF2_Buffer_WriteUint32( buffer, attributeList->capacity );

    for ( otf2_attribute* attr = attributeList->head; attr; attr = attr->next )
    {
        /* Write attribute id to the trace */
        OTF2_Buffer_WriteUint32( buffer, attr->attribute_id );

        /* Write the type to the trace */
        OTF2_Buffer_WriteUint8( buffer, attr->type_id );

        switch ( attr->type_id )
        {
            @otf2 for entry in Type.entries if entry.type:
            case @@entry.name@@:
                OTF2_Buffer_Write@@entry.type.otf2@@( buffer,
                                                      attr->value.@@entry.type.lower@@ );
                break;

            @otf2 endfor
            @otf2 for entry in MappingType.entries:
            @otf2 with def = entry.def
            case @@Type.prefix@@@@entry.suffix@@:
                /* @@Type.prefix@@@@entry.suffix@@ attributes reference a @@def.siblings|join(', or a ')@@ definition */
                OTF2_Buffer_Write@@def.self_attr.type.otf2@@( buffer, attr->value.@@def.ref_lname@@Ref );
                break;

            @otf2 endwith
            @otf2 endfor
            default:
                /*
                 * When getting an unknown type, we write it nevertheless as an
                 * uint64, because we assume that future additions will always be
                 * definition references and they are always of compressed types.
                 */
                OTF2_Buffer_WriteUint64( buffer, attr->value.uint64 );
                break;
        }
    }

    otf2_attribute_list_remove_all_attributes( attributeList );

    /* Write real record length */
    return OTF2_Buffer_WriteFinalRecordLength( buffer, data_size_of_attr_list );
}

