#!/bin/sh

##
## This file is part of the Score-P software (http://www.score-p.org)
##
## Copyright (c) 2013,
## Technische Universitaet Dresden, Germany
##
## This software may be modified and distributed under the terms of
## a BSD-style license.  See the COPYING file in the package base
## directory for details.
##

set -e

cleanup()
{
    rm -rf test-trace-parallel-sion-*
}
trap cleanup EXIT
cleanup

getsubset()
{
    n=$1
    i=$2
    bit=1
    ns=0
    is=""
    for b in $(seq 0 $(($n - 1)))
    do
        if test $(((1 << b) & i)) -gt 0
        then
            ns=$((ns + 1))
            is="$is $b"
        fi
    done
    echo $is
}

tests=0
successes=0
fails=0
failed=
failed_sep=
nl='
'
n=0
for i in 1 2 3 4
do
    # 0 equals 1 file, but without sub-communicator
    for f in $(seq 0 $i)
    do
        t=test-trace-parallel-sion-$i.$f

        tests=$((tests + 1))
        n=$((n + 1))
        writer_cmd="$failed_sep./test-writer -nt $i $t 1 $f"
        test ${V-0} -eq 1 && echo "$n - Writing SION trace with $i rank(s) into $f file(s) (w/ local defs): $writer_cmd"
        if $VALGRIND $writer_cmd
        then
            successes=$((successes + 1))
        else
            fails=$((fails + 1))
            failed="$failedfailed_sep$writer_cmd"
            failed_sep=$nl
            continue
        fi
        m=0
        for j in $(seq 1 $i) $((2 * i))
        do
            for k in $(seq 0 $(((1 << $i) - 1)))
            do
                tests=$((tests + 1))
                m=$((m + 1))
                reader_cmd="./test-reader -nt $j $t.otf2 0 $(getsubset $i $k)"
                test ${V-0} -eq 1 && echo "$n.$m - Reading with $j rank(s) $k out of $i rank(s) out of $f files (w/o sub-groups): $reader_cmd"
                if $VALGRIND $reader_cmd
                then
                    successes=$((successes + 1))
                else
                    fails=$((fails + 1))
                    failed="$failed$writer_cmd$nl$reader_cmd"
                    writer_cmd=
                fi

                tests=$((tests + 1))
                m=$((m + 1))
                reader_cmd="./test-reader -nt $j $t.otf2 1 $(getsubset $i $k)"
                test ${V-0} -eq 1 && echo "$n.$m - Reading with $j rank(s) $k out of $i rank(s) out of $f files (w/ sub-groups): $reader_cmd"
                if $VALGRIND $reader_cmd
                then
                    successes=$((successes + 1))
                else
                    fails=$((fails + 1))
                    failed="$failed$writer_cmd$nl$reader_cmd"
                    writer_cmd=
                fi

                tests=$((tests + 1))
                m=$((m + 1))
                reader_cmd="./test-reader -nt $j $t.otf2 2 $(getsubset $i $k)"
                test ${V-0} -eq 1 && echo "$n.$m - Reading with $j rank(s) $k out of $i rank(s) out of $f files (w/ serial): $reader_cmd"
                if $VALGRIND $reader_cmd
                then
                    successes=$((successes + 1))
                else
                    fails=$((fails + 1))
                    failed="$failed$writer_cmd$nl$reader_cmd"
                    writer_cmd=
                fi
            done
        done

        # test without local def files
        t_nold="$t-nold"
        tests=$((tests + 1))
        writer_cmd="$failed_sep./test-writer -nt $i $t_nold 0 $f"
        test ${V-0} -eq 1 && echo "$n - Writing SION trace with $i rank(s) into $f file(s) (w/o local defs): $writer_cmd"
        if $VALGRIND $writer_cmd
        then
            successes=$((successes + 1))
        else
            fails=$((fails + 1))
            failed="$failedfailed_sep$writer_cmd"
            failed_sep=$nl
            continue
        fi

        # read the trace once
        j=$i
        k=$(((1 << $i) - 1))
        tests=$((tests + 1))
        m=$((m + 1))
        reader_cmd="./test-reader -nt $j $t_nold.otf2 0 $(getsubset $i $k)"
        test ${V-0} -eq 1 && echo "$n.$m - Reading with $j rank(s) $k out of $i rank(s) out of $f files (w/o sub-groups): $reader_cmd"
        if $VALGRIND $reader_cmd
        then
            successes=$((successes + 1))
        else
            fails=$((fails + 1))
            failed="$failed$writer_cmd$nl$reader_cmd"
            writer_cmd=
        fi

        cleanup
    done
done

echo "Success: $successes / $tests"
test $fails -eq 0 && exit 0
echo "Fails: $fails / $tests"
test ${V-0} -eq 1 && echo "$failed"
exit 1
